#include "yolov5det.h"

#include "model.h"
#include <fstream>

using namespace nvinfer1;
using namespace std;

yolov5det::yolov5det()
{

}

int yolov5det::createngine(std::string &wts_name, std::string &engine_name, char s_value){
    bool is_p6 = false;
    float gd = 0.0f, gw = 0.0f;

    if (s_value == 'n') {
      gd = 0.33;
      gw = 0.25;
    } else if (s_value == 's') {
      gd = 0.33;
      gw = 0.50;
    } else if (s_value == 'm') {
      gd = 0.67;
      gw = 0.75;
    } else if (s_value == 'l') {
      gd = 1.0;
      gw = 1.0;
    } else if (s_value == 'x') {
      gd = 1.33;
      gw = 1.25;
    }

    serialize_engine(kBatchSize, is_p6, gd, gw, wts_name, engine_name);

    return 1;
}

yolov5det::~yolov5det(){
    if(cpu_output_buffer != nullptr){
        cudaStreamDestroy(stream);
        CUDA_CHECK(cudaFree(gpu_buffers[0]));
        CUDA_CHECK(cudaFree(gpu_buffers[1]));
        delete[] cpu_output_buffer;
        cuda_preprocess_destroy();
        // Destroy the engine
        context->destroy();
        engine->destroy();
        runtime->destroy();
    }
}

int yolov5det::setengine(std::string& engine_name){
    // Deserialize the engine from file
    deserialize_engine(engine_name, &runtime, &engine, &context);
    CUDA_CHECK(cudaStreamCreate(&stream));

    // Init CUDA preprocessing
    cuda_preprocess_init(kMaxInputImageSize);

    // Prepare cpu and gpu buffers
    prepare_buffers(engine, &gpu_buffers[0], &gpu_buffers[1], &cpu_output_buffer);
    return 1;
}

std::vector<std::vector<Detection>> yolov5det::detect(cv::Mat& image){
    std::vector<cv::Mat> img_batch;
    img_batch.push_back(image);

    // Preprocess
    cuda_batch_preprocess(img_batch, gpu_buffers[0], kInputW, kInputH, stream);
    // Run inference
    infer(*context, stream, (void**)gpu_buffers, cpu_output_buffer, kBatchSize);
    // NMS
    std::vector<std::vector<Detection>> res_batch;
    batch_nms(res_batch, cpu_output_buffer, img_batch.size(), kOutputSize, kConfThresh, kNmsThresh);

    return res_batch;
}


void yolov5det::serialize_engine(unsigned int max_batchsize, bool& is_p6, float& gd, float& gw, std::string& wts_name, std::string& engine_name) {
  // Create builder
  IBuilder* builder = createInferBuilder(gLogger);
  IBuilderConfig* config = builder->createBuilderConfig();

  // Create model to populate the network, then set the outputs and create an engine
  ICudaEngine *engine = nullptr;
  if (is_p6) {
    engine = build_det_p6_engine(max_batchsize, builder, config, DataType::kFLOAT, gd, gw, wts_name,kInputH ,kInputW,kNumClass,kMaxNumOutputBbox,kInputTensorName,kOutputTensorName);
  } else {
    engine = build_det_engine(max_batchsize, builder, config, DataType::kFLOAT, gd, gw, wts_name,kInputH ,kInputW,kNumClass,kMaxNumOutputBbox,kInputTensorName,kOutputTensorName);
  }
  assert(engine != nullptr);

  // Serialize the engine
  IHostMemory* serialized_engine = engine->serialize();
  assert(serialized_engine != nullptr);

  // Save engine to file
  std::ofstream p(engine_name, std::ios::binary);
  if (!p) {
    std::cerr << "Could not open plan output file" << std::endl;
    assert(false);
  }
  p.write(reinterpret_cast<const char*>(serialized_engine->data()), serialized_engine->size());

  // Close everything down
  engine->destroy();
  builder->destroy();
  config->destroy();
  serialized_engine->destroy();
}

void yolov5det::deserialize_engine(std::string& engine_name, IRuntime** runtime, ICudaEngine** engine, IExecutionContext** context) {
  std::ifstream file(engine_name, std::ios::binary);
  if (!file.good()) {
    std::cerr << "read " << engine_name << " error!" << std::endl;
    assert(false);
  }
  size_t size = 0;
  file.seekg(0, file.end);
  size = file.tellg();
  file.seekg(0, file.beg);
  char* serialized_engine = new char[size];
  assert(serialized_engine);
  file.read(serialized_engine, size);
  file.close();

  *runtime = createInferRuntime(gLogger);
  assert(*runtime);
  *engine = (*runtime)->deserializeCudaEngine(serialized_engine, size);
  assert(*engine);
  *context = (*engine)->createExecutionContext();
  assert(*context);
  delete[] serialized_engine;
}

void yolov5det::prepare_buffers(nvinfer1::ICudaEngine* engine, float** gpu_input_buffer, float** gpu_output_buffer, float** cpu_output_buffer){
    assert(engine->getNbBindings() == 2);
    // In order to bind the buffers, we need to know the names of the input and output tensors.
    // Note that indices are guaranteed to be less than IEngine::getNbBindings()
    const int inputIndex = engine->getBindingIndex(kInputTensorName);
    const int outputIndex = engine->getBindingIndex(kOutputTensorName);
    assert(inputIndex == 0);
    assert(outputIndex == 1);
    // Create GPU buffers on device
    CUDA_CHECK(cudaMalloc((void**)gpu_input_buffer, kBatchSize * 3 * kInputH * kInputW * sizeof(float)));
    CUDA_CHECK(cudaMalloc((void**)gpu_output_buffer, kBatchSize * kOutputSize * sizeof(float)));

    *cpu_output_buffer = new float[kBatchSize * kOutputSize];
    std::cout<<"kBatchSize * kOutputSize:"<<kBatchSize * kOutputSize<<endl;
}

void yolov5det::infer(IExecutionContext& context, cudaStream_t& stream, void** gpu_buffers, float* output, int batchsize) {
  context.enqueue(batchsize, gpu_buffers, stream, nullptr);
  CUDA_CHECK(cudaMemcpyAsync(output, gpu_buffers[1], batchsize * kOutputSize * sizeof(float), cudaMemcpyDeviceToHost, stream));
  cudaStreamSynchronize(stream);
}
